from __future__ import annotations

import contextlib
import io
import logging
import os
import shlex
import shutil
import subprocess
import sys
import traceback
import warnings
from pathlib import Path
from typing import Any, Callable, Iterator, Sequence, Union
from unittest import mock

import pytest

if sys.version_info < (3, 10):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata

_StrOrPath = Union[str, os.PathLike]
"""A command line argument type as a str or path."""

_Command = Union[_StrOrPath, Sequence[_StrOrPath]]
"""A command-like type compatible with subprocess.run."""

StreamMock = io.StringIO


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('console-scripts')
    group.addoption(
        '--script-launch-mode',
        metavar='inprocess|subprocess|both',
        action='store',
        dest='script_launch_mode',
        default=None,
        help='how to run python scripts under test (default: inprocess)'
    )
    group.addoption(
        '--hide-run-results',
        action='store_true',
        dest='hide_run_results',
        default=False,
        help="don't print out script run results on failures or when "
             'output capturing is disabled'
    )
    parser.addini(
        'script_launch_mode',
        'how to run python scripts under test (inprocess|subprocess|both)'
    )


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line(
        'markers',
        'script_launch_mode: how to run python scripts under test '
        '(inprocess|subprocess|both)',
    )


def _get_mark_mode(metafunc: pytest.Metafunc) -> str | None:
    """Return launch mode as indicated by test function marker or None."""
    marker = metafunc.definition.get_closest_marker('script_launch_mode')
    if marker:
        return str(marker.args[0])
    return None


def _is_nonexecutable_python_file(command: _StrOrPath) -> bool:
    """Check if `command` is a Python file with no executable mode set."""
    command = Path(command)
    mode = command.stat().st_mode
    if mode & os.X_OK:
        return False
    return command.suffix == '.py'


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Parametrize script_launch_mode fixture.

    Checks the configuration sources in this order:
    - `script_launch_mode` mark on the test,
    - `--script-launch-mode` option,
    - `script_launch_mode` configuration option in [pytest] section of the
      pyest config file.

    This process yields a value that can be one of:
    - "inprocess" -- The script will be run via loading its main function
      into the test runner process and mocking the environment.
    - "subprocess" -- The script will be run via `subprocess` module.
    - "both" -- The test will be run twice: once in inprocess mode and once
      in subprocess mode.
    - None -- Same as "inprocess".
    """
    if 'script_launch_mode' not in metafunc.fixturenames:
        return

    mark_mode = _get_mark_mode(metafunc)
    option_mode = metafunc.config.option.script_launch_mode
    config_mode = metafunc.config.getini('script_launch_mode')

    mode = mark_mode or option_mode or config_mode or 'inprocess'

    if mode in {'inprocess', 'subprocess'}:
        metafunc.parametrize('script_launch_mode', [mode], indirect=True)
    elif mode == 'both':
        metafunc.parametrize('script_launch_mode', ['inprocess', 'subprocess'],
                             indirect=True)
    else:
        raise ValueError(f'Invalid script launch mode: {mode}')


class RunResult:
    """Result of running a script."""

    def __init__(
        self, returncode: int, stdout: str, stderr: str, print_result: bool
    ) -> None:
        self.success = returncode == 0
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        if print_result:
            self.print()

    def print(self) -> None:
        print('# Script return code:', self.returncode)
        print('# Script stdout:', self.stdout, sep='\n')
        print('# Script stderr:', self.stderr, sep='\n')


def _handle_command_args(
    command: _Command,
    *args: _StrOrPath,
    shell: bool = False,
) -> Sequence[_StrOrPath]:
    """Return command arguments in a consistent list format.

    If shell=True then this function tries to mimic local shell execution.
    """
    if shell:
        if args or not isinstance(command, (str, os.PathLike)):
            command = subprocess.list2cmdline(
                str(arg)
                for arg in _handle_command_args(command, *args, shell=False)
            )
        command = shlex.split(str(command), posix=os.name == 'posix')
        args = ()

    if args:
        warnings.warn(
            'script_runner commands should be passed as a single sequence,'
            ' not as multiple arguments.'
            '\nReplace `script_runner.run(a, b, c)` calls with'
            ' `script_runner.run([a, b, c])`',
            DeprecationWarning,
        )
        if not isinstance(command, (str, os.PathLike)):
            return [*command, *args]
        return [command, *args]
    if isinstance(command, (str, os.PathLike)):
        return [command]
    return command


@contextlib.contextmanager
def _patch_environ(new_environ: dict[str, str] | None) -> Iterator[None]:
    """Replace the environment for the duration of a context."""
    if new_environ is None:
        yield
        return
    old_environ = os.environ.copy()
    os.environ.clear()
    os.environ.update(new_environ)
    yield
    os.environ.clear()
    os.environ.update(old_environ)


@contextlib.contextmanager
def _chdir_context(new_dir: _StrOrPath | None) -> Iterator[None]:
    """Replace the current directory for the duration of a context."""
    if new_dir is None:
        yield
        return
    old_cwd = os.getcwd()
    os.chdir(new_dir)
    yield
    os.chdir(old_cwd)


@contextlib.contextmanager
def _push_and_reset_logger() -> Iterator[None]:
    """Do a very basic reset of the root logger and restore its config on exit.

    This allows scripts to call logging.basicConfig(...) and have
    it work as expected. It might not work for more sophisticated logging
    setups but it's simple and covers the basic usage whereas implementing
    a comprehensive fix is impossible in a compatible way.
    """
    logger = logging.getLogger()
    old_handlers = logger.handlers
    old_disabled = logger.disabled
    old_level = logger.level
    logger.handlers = []
    logger.disabled = False
    logger.setLevel(logging.NOTSET)
    yield
    # Restore logger to previous configuration
    logger.handlers = old_handlers
    logger.disabled = old_disabled
    logger.setLevel(old_level)


class ScriptRunner:
    """Fixture for running python scripts under test."""

    def __init__(
        self, launch_mode: str,
        rootdir: _StrOrPath,
        print_result: bool = True
    ) -> None:
        assert launch_mode in {'inprocess', 'subprocess'}
        self.launch_mode = launch_mode
        self.print_result = print_result
        self.rootdir = rootdir

    def __repr__(self) -> str:
        return f'<ScriptRunner {self.launch_mode}>'

    def run(
        self,
        command: _Command,
        *arguments: _StrOrPath,
        print_result: bool | None = None,
        shell: bool = False,
        cwd: _StrOrPath | None = None,
        env: dict[str, str] | None = None,
        stdin: io.IOBase | None = None,
        check: bool = False,
        **options: Any,
    ) -> RunResult:
        if print_result is None:
            print_result = self.print_result

        if print_result:
            print('# Running console script:', command, *arguments)

        if self.launch_mode == 'inprocess':
            run_function = self.run_inprocess
        else:
            run_function = self.run_subprocess
        return run_function(
            command,
            *arguments,
            print_result=print_result,
            shell=shell,
            cwd=cwd,
            env=env,
            stdin=stdin,
            check=check,
            **options,
        )

    @staticmethod
    def _locate_script(
        command: _StrOrPath,
        *,
        cwd: _StrOrPath | None,
        env: dict[str, str] | None,
    ) -> Path:
        """Locate script in PATH or in current directory."""
        script_path = shutil.which(
            command,
            path=env.get('PATH', None) if env is not None else None,
        )
        if script_path is not None:
            return Path(script_path)

        cwd = cwd if cwd is not None else os.getcwd()
        return Path(cwd, command).resolve(strict=True)

    @classmethod
    def _load_script(
        cls,
        command: _StrOrPath,
        *,
        cwd: _StrOrPath | None,
        env: dict[str, str] | None,
    ) -> Callable[[], int | None]:
        """Load target script via entry points or compile/exec."""
        if isinstance(command, str):
            entry_points = tuple(
                importlib_metadata.entry_points(
                    group='console_scripts', name=command
                )
            )
            if entry_points:
                def console_script() -> int | None:
                    s: Callable[[], int | None] = entry_points[0].load()
                    return s()
                return console_script

        script_path = cls._locate_script(command, cwd=cwd, env=env)

        def exec_script() -> int:
            with open(script_path, 'rt', encoding='utf-8') as script:
                compiled = compile(script.read(), str(script), 'exec', flags=0)
                exec(compiled, {'__name__': '__main__'})
            return 0

        return exec_script

    @classmethod
    def run_inprocess(
        cls,
        command: _Command,
        *arguments: _StrOrPath,
        shell: bool = False,
        cwd: _StrOrPath | None = None,
        env: dict[str, str] | None = None,
        print_result: bool = True,
        stdin: io.IOBase | None = None,
        check: bool = False,
        **options: Any,
    ) -> RunResult:
        for key in options:
            warnings.warn(
                f'Keyword argument {key!r} was ignored.'
                '\nConsider using subprocess mode or raising an issue.'
            )
        cmd_args = _handle_command_args(command, *arguments, shell=shell)
        script = cls._load_script(cmd_args[0], cwd=cwd, env=env)
        cmd_args = [str(cmd) for cmd in cmd_args]
        stdin_stream = stdin if stdin is not None else StreamMock()
        stdout_stream = StreamMock()
        stderr_stream = StreamMock()
        with contextlib.ExitStack() as stack:
            stack.enter_context(mock.patch('sys.stdin', new=stdin_stream))
            stack.enter_context(contextlib.redirect_stdout(stdout_stream))
            stack.enter_context(contextlib.redirect_stderr(stderr_stream))
            stack.enter_context(mock.patch('sys.argv', new=cmd_args))
            stack.enter_context(_push_and_reset_logger())
            stack.enter_context(_patch_environ(env))
            stack.enter_context(_chdir_context(cwd))

            try:
                returncode = script()
            except SystemExit as exc:
                returncode = 1
                if isinstance(exc.code, str):
                    stderr_stream.write(f'{exc}\n')
                    returncode = 1
                else:
                    returncode = exc.code
            except Exception:
                returncode = 1
                try:
                    et, ev, tb = sys.exc_info()
                    assert tb
                    # Hide current frame from the stack trace.
                    traceback.print_exception(et, ev, tb.tb_next)
                finally:
                    del tb

        result = RunResult(
            returncode or 0,  # None also means success
            stdout_stream.getvalue(),
            stderr_stream.getvalue(),
            print_result,
        )

        if check and returncode:
            raise subprocess.CalledProcessError(
                returncode,
                cmd_args,
                result.stdout,
                result.stderr,
            )

        return result

    @classmethod
    def run_subprocess(
        cls,
        command: _Command,
        *arguments: _StrOrPath,
        print_result: bool = True,
        shell: bool = False,
        cwd: _StrOrPath | None = None,
        env: dict[str, str] | None = None,
        stdin: io.IOBase | None = None,
        check: bool = False,
        universal_newlines: bool = True,
        **options: Any,
    ) -> RunResult:
        stdin_input: str | bytes | None = None
        if stdin is not None:
            stdin_input = stdin.read()

        script_path = cls._locate_script(_handle_command_args(
            command, *arguments, shell=shell)[0], cwd=cwd, env=env
        )
        if arguments:
            command = _handle_command_args(command, *arguments, shell=shell)

        if _is_nonexecutable_python_file(script_path):
            command = _handle_command_args(command, shell=shell)
            command = [sys.executable or 'python', *command]

        if sys.version_info < (3, 8):
            if isinstance(command, os.PathLike):
                command = [command]
            if not isinstance(command, str):
                command = [
                    str(Path(arg)) if isinstance(arg, os.PathLike) else arg
                    for arg in command
                ]

        cp = subprocess.run(
            command,
            input=stdin_input,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
            cwd=cwd,
            env=env,
            check=check,
            universal_newlines=universal_newlines,
            **options,
        )
        return RunResult(cp.returncode, cp.stdout, cp.stderr, print_result)


@pytest.fixture
def script_launch_mode(request: pytest.FixtureRequest) -> str:
    return str(request.param)


@pytest.fixture
def script_cwd(tmp_path: Path) -> Path:
    work_dir = tmp_path / 'script-cwd'
    work_dir.mkdir()
    return work_dir


@pytest.fixture
def script_runner(
    request: pytest.FixtureRequest, script_cwd: Path, script_launch_mode: str
) -> ScriptRunner:
    print_result = not request.config.getoption('--hide-run-results')
    return ScriptRunner(script_launch_mode, script_cwd, print_result)



"""Test running of scripts with various modes and options."""
from __future__ import annotations

import importlib
import io
import os
import sys
from pathlib import Path
from subprocess import CalledProcessError
from types import ModuleType
from typing import Any
from unittest import mock

import pytest

from pytest_console_scripts import ScriptRunner


@pytest.fixture(params=['inprocess', 'subprocess'])
def launch_mode(request: pytest.FixtureRequest) -> str:
    """Launch mode: inprocess|subprocess."""
    return str(request.param)


@pytest.fixture()
def console_script(tmp_path: Path) -> Path:
    """Python script to use in tests."""
    script = tmp_path / 'script.py'
    script.write_text('#!/usr/bin/env python\nprint("foo")')
    return script


@pytest.mark.script_launch_mode('both')
def test_not_installed(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    result = script_runner.run(str(console_script))
    assert result.success
    assert result.stdout == 'foo\n'
    assert result.stderr == ''


@pytest.mark.xfail(
    sys.platform == "win32",
    reason="Windows does not treat Python scripts as executables."
)
@pytest.mark.script_launch_mode('both')
def test_elsewhere_in_the_path(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    console_script.chmod(0o777)
    env = os.environ.copy()
    env["PATH"] = f"{console_script.parent}{os.pathsep}{env['PATH']}"
    result = script_runner.run(console_script.name, env=env)
    assert result.success
    assert result.stdout == 'foo\n'
    assert result.stderr == ''


@pytest.mark.script_launch_mode('both')
def test_run_pytest(
    tmp_path: Path,
    console_script: Path,
    script_runner: ScriptRunner,
    launch_mode: str
) -> None:
    console_script.write_text('import os;print(os.getpid())')
    test = tmp_path / f'test_{launch_mode}.py'
    compare = '==' if launch_mode == 'inprocess' else '!='
    test.write_text(
        f"""
import os
def test_script(script_runner):
    result = script_runner.run(R'''{console_script}''')
    assert result.success
    assert result.stdout {compare} str(os.getpid()) + '\\n'
    assert result.stderr == ''
        """
    )

    # Here we're testing two things:
    #
    # - pytest is a Python script that's installed in the test environment, so
    #   we'll use `script_runner` fixture to run it -- this tests execution of
    #   installed scripts from the path.
    # - The pytest that we run will run a test that uses `script_runner`
    #   fixture to run another script. We're going to pass --script-launch-mode
    #   option to pytest and will check that the execution of the inner script
    #   is performed in accordance with its value.
    #
    # We're also testing all 4 combinations of inprocess/subprocess modes for
    # inner and outer script runners.

    result = script_runner.run(
        'pytest',
        str(test),
        '--script-launch-mode=' + launch_mode,
    )
    assert result.success


@pytest.mark.script_launch_mode('inprocess')
def test_return_None(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    """Check that entry point function returning None is counted as success."""
    # Many console_scripts entry point functions return 0 on success but not
    # all of them do. Returning `None` is also allowed and would be translated
    # to return code 0 when run normally via wrapper. This test checks that we
    # handle this case properly in inprocess mode.
    console_script.write_text(
        """
import sys
print("Foo")
sys.exit(None)
"""
    )
    result = script_runner.run(str(console_script))
    assert result.success
    assert 'Foo' in result.stdout


@pytest.mark.script_launch_mode('inprocess')
def test_return_code_uncommon(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    """Check uncommon return codes."""
    console_script.write_text(
        """
import sys
sys.exit(2)
"""
    )
    assert script_runner.run(str(console_script)).returncode == 2


@pytest.mark.script_launch_mode('both')
def test_abnormal_exit(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    console_script.write_text('import sys;sys.exit("boom")')
    result = script_runner.run(str(console_script))
    assert not result.success
    assert result.stdout == ''
    assert result.stderr == 'boom\n'


@pytest.mark.script_launch_mode('both')
def test_exception(console_script: Path, script_runner: ScriptRunner) -> None:
    console_script.write_text('raise TypeError("boom")')
    result = script_runner.run(str(console_script))
    assert not result.success
    assert result.stdout == ''
    assert 'TypeError: boom' in result.stderr


def test_cwd(
    console_script: Path,
    script_runner: ScriptRunner,
    tmp_path: Path,
) -> None:
    """Script starts in dir given by cwd arg and cwd changes are contained."""
    dir1 = tmp_path / 'dir1'
    dir1.mkdir()
    dir2 = tmp_path / 'dir2'
    dir2.mkdir()
    console_script.write_text(
        f"""
import os
print(os.getcwd())
os.chdir(R'''{dir2}''')
print(os.getcwd())
        """
    )
    mydir = os.getcwd()
    result = script_runner.run(str(console_script), cwd=str(dir1))
    assert result.success
    assert result.stdout == f'{dir1}\n{dir2}\n'
    assert os.getcwd() == mydir


@pytest.mark.script_launch_mode('both')
def test_env(console_script: Path, script_runner: ScriptRunner) -> None:
    """Script receives environment and env changes don't escape to test."""
    console_script.write_text(
        """
import os
print(os.environ['FOO'])
os.environ['FOO'] = 'baz'
        """
    )
    env = os.environ.copy()
    env['FOO'] = 'bar'
    result = script_runner.run(str(console_script), env=env)
    assert result.success
    assert result.stdout == 'bar\n'
    assert 'FOO' not in os.environ


@pytest.mark.script_launch_mode('both')
def test_stdin(console_script: Path, script_runner: ScriptRunner) -> None:
    console_script.write_text(
        """
import sys
for line in sys.stdin:
    sys.stdout.write('simon says ' + line)
    sys.stderr.write('error says ' + line)
        """
    )
    stdin = io.StringIO('foo\nbar')
    result = script_runner.run(str(console_script), stdin=stdin)
    assert result.success
    assert result.stdout == 'simon says foo\nsimon says bar'
    assert result.stderr == 'error says foo\nerror says bar'


def test_logging(console_script: Path, script_runner: ScriptRunner) -> None:
    """Test that the script can perform logging initialization."""
    console_script.write_text(
        """
import logging, sys
logging.basicConfig(stream=sys.stderr, level=logging.INFO)
logging.debug('hidden')
logging.info('shown')
        """
    )
    result = script_runner.run(str(console_script))
    assert result.success
    assert result.stderr == 'INFO:root:shown\n'


@pytest.mark.parametrize('fail', [True, False])
def test_print_stdio_on_error(
    console_script: Path,
    script_runner: ScriptRunner,
    tmp_path: Path,
    fail: bool,
    launch_mode: str,
) -> None:
    """Output of the script is printed when the test fails."""
    console_script.write_text('print("12345")\nraise Exception("54321")')
    test = tmp_path / f'test_{fail}_{launch_mode}.py'
    test.write_text(
        f"""
def test_fail(script_runner):
    ret = script_runner.run(R'''{console_script}''', 'arg')
    assert ret.success is {fail}
        """
    )
    result = script_runner.run(
        'pytest',
        str(test),
        '--script-launch-mode=' + launch_mode,
    )
    assert result.success != fail
    if fail:
        assert (f'# Running console script: {console_script} arg\n'
                in result.stdout)
        assert '# Script return code: 1\n' in result.stdout
        assert '# Script stdout:\n12345\n' in result.stdout
        assert '# Script stderr:\nTraceback' in result.stdout
        assert 'Exception: 54321' in result.stdout
    else:
        assert '# Running console script' not in result.stdout
        assert '12345' not in result.stdout
        assert '54321' not in result.stdout


@pytest.mark.script_launch_mode('inprocess')
def test_mocking(
    console_script: Path,
    script_runner: ScriptRunner,
    monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test mocking in of console scripts (in-process mode only).

    Note: we can't mock objects in the script itself because it will not be
    imported via normal import system but we can mock anything in the modules
    that the script imports.

    """
    console_script.write_text(
        """
import os
print(os.path.basename('foo'))
        """
    )
    monkeypatch.setattr(os.path, 'basename', lambda foo: 'bar')
    result = script_runner.run(str(console_script))
    assert result.success
    assert result.stdout == 'bar\n'


def test_hide_run_result_arg(
    tmp_path: Path, console_script: Path, script_runner: ScriptRunner
) -> None:
    """Disable printing of the RunResult to stdout with print_result=False."""
    console_script.write_text('print("the answer is 42")')
    test = tmp_path / 'test_hrra.py'
    test.write_text(
        f"""
import pytest

@pytest.mark.script_launch_mode('both')
def test_script(script_runner):
    script_runner.run(R'''{console_script}''', print_result=False)
        """
    )
    result = script_runner.run('pytest', '-s', str(test))
    assert result.success
    assert 'the answer is 42' not in result.stdout
    assert 'Running console script' not in result.stdout


def test_hide_run_result_opt(
    tmp_path: Path, console_script: Path, script_runner: ScriptRunner
) -> None:
    """Disable printing of the RunResult to stdout with print_result=False."""
    console_script.write_text('print("the answer is 42")')
    test = tmp_path / 'test_hrro.py'
    test.write_text(
        f"""
import pytest

@pytest.mark.script_launch_mode('both')
def test_script(script_runner):
    script_runner.run(R'''{console_script}''')
        """
    )
    result = script_runner.run('pytest', '-s', '--hide-run-results', str(test))
    assert result.success
    assert 'the answer is 42' not in result.stdout
    assert 'Running console script' not in result.stdout


class MockEntryPoint:
    module: ModuleType

    def __init__(self, exec_path: str | Path):
        self.exec_path = exec_path

    def load(self) -> Any:
        base, module = os.path.split(self.exec_path)
        module_name, _ = os.path.splitext(module)
        sys.path.append(base)
        self.module = importlib.import_module(module_name)
        sys.path.pop(-1)
        return self.module.run


@pytest.mark.script_launch_mode('inprocess')
def test_global_logging(
    tmp_path: Path, console_script: Path, script_runner: ScriptRunner
) -> None:
    """Load global values when executing from importlib.metadata"""
    test = tmp_path / 'test_entry_point.py'
    test.write_text(
        """
import logging

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


def run() -> None:
    LOGGER.debug('DEBUG')
    LOGGER.info('INFO')
    LOGGER.warning('WARNING')
        """
    )

    if sys.version_info < (3, 10):
        patched_func = 'importlib_metadata.entry_points'
    else:
        patched_func = 'importlib.metadata.entry_points'

    with mock.patch(
        patched_func,
        mock.MagicMock(return_value=[MockEntryPoint(str(test))]),
    ):
        result = script_runner.run(str(console_script))
        assert result.success
        assert 'INFO:test_entry_point:INFO\n' in result.stderr
        assert 'DEBUG\n' not in result.stderr


@pytest.mark.script_launch_mode('both')
def test_shell(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    console_script.chmod(0o777)
    result = script_runner.run(
        f"{console_script} --test", shell=True, check=True
    )
    assert result.stdout == 'foo\n'
    assert result.stderr == ''
    result = script_runner.run(
        [str(console_script), "--test"], shell=True, check=True
    )
    assert result.stdout == 'foo\n'
    assert result.stderr == ''


@pytest.mark.script_launch_mode('both')
def test_deprecated_args(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    if (
        script_runner.launch_mode == 'subprocess'
        and sys.platform == 'win32'
        and sys.version_info < (3, 8)
    ):
        pytest.xfail("PathLike's might not be supported this far back")
    console_script.write_text(
        """
import sys
print(sys.argv[1:])
        """
    )
    result = script_runner.run(console_script, 'A', 'B', check=True)
    assert result.stdout == "['A', 'B']\n"
    result = script_runner.run([console_script, 'C'], 'D', check=True)
    assert result.stdout == "['C', 'D']\n"


@pytest.mark.script_launch_mode('both')
def test_check(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    console_script.write_text("""import sys; sys.exit(1)""")
    with pytest.raises(CalledProcessError, match='.*non-zero exit status 1'):
        script_runner.run(str(console_script), check=True)


@pytest.mark.script_launch_mode('both')
def test_ignore_universal_newlines(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    result = script_runner.run(
        str(console_script), check=True, universal_newlines=True
    )
    assert result.stdout == 'foo\n'
    assert result.stderr == ''


@pytest.mark.script_launch_mode('subprocess')
def test_disable_universal_newlines(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    result = script_runner.run(
        str(console_script), check=True, universal_newlines=False
    )
    assert isinstance(result.stdout, bytes)
    assert isinstance(result.stderr, bytes)
    assert result.stdout.strip() == b'foo'
    assert result.stderr == b''


@pytest.mark.script_launch_mode('both')
def test_run_path(
    console_script: Path, script_runner: ScriptRunner
) -> None:
    result = script_runner.run(console_script, check=True)
    assert result.stdout == 'foo\n'
    assert result.stderr == ''
    console_script.chmod(0o777)
    result = script_runner.run(console_script, check=True)
    assert result.stdout == 'foo\n'
    assert result.stderr == ''



from __future__ import annotations

import pytest


@pytest.fixture(params=[None, 'inprocess', 'subprocess', 'both'])
def launch_mode_conf(request: pytest.FixtureRequest) -> str | None:
    """Configured launch mode (None|'inprocess'|'subprocess'|'both')."""
    assert request.param is None or isinstance(request.param, str)
    return request.param


@pytest.fixture
def launch_modes(launch_mode_conf: str | None) -> set[str]:
    """Set of launch modes in which the tests will actually be run.

    The value of this fixture depends on the value of `launch_mode_conf`:
    - 'inprocess'  -> {'inprocess'}
    - 'subprocess' -> {'subprocess'}
    - 'both'       -> {'inprocess', 'subprocess'}
    - None         -> {'inprocess'}
    """
    if launch_mode_conf == 'both':
        return {'inprocess', 'subprocess'}
    if launch_mode_conf is not None:
        return {launch_mode_conf}
    return {'inprocess'}


class RunTest:
    def __init__(self, testdir: pytest.Testdir) -> None:
        self.testdir = testdir

    def __call__(
        self,
        script: str,
        passed: int = 1,
        skipped: int = 0,
        failed: int = 0,
        launch_mode_conf: str | None = None
    ) -> pytest.RunResult:
        self.testdir.makepyfile(script)
        args = []
        if launch_mode_conf is not None:
            args.append('--script-launch-mode=' + launch_mode_conf)
        result = self.testdir.runpytest(*args)
        print('\n'.join(['pytest stdout:'] + result.outlines +
                        ['pytest stderr:'] + result.errlines))
        result.assert_outcomes(passed=passed, skipped=skipped, failed=failed)
        return result


@pytest.fixture
def run_test(testdir: pytest.Testdir) -> RunTest:
    return RunTest(testdir)


CHECK_LAUNCH_MODE = """
def test_both(script_runner, accumulator=set()):
    assert script_runner.launch_mode in {}
    assert script_runner.launch_mode not in accumulator
    accumulator.add(script_runner.launch_mode)
"""


def test_command_line_option(
    run_test: RunTest, launch_mode_conf: str | None, launch_modes: set[str]
) -> None:
    run_test(
        CHECK_LAUNCH_MODE.format(launch_modes),
        passed=len(launch_modes),
        launch_mode_conf=launch_mode_conf
    )


def test_config_option(
    run_test: RunTest,
    testdir: pytest.Testdir,
    launch_mode_conf: str | None,
    launch_modes: set[str],
) -> None:
    if launch_mode_conf is not None:
        testdir.makeini(f"""
            [pytest]
            script_launch_mode = {launch_mode_conf}
        """)

    run_test(
        CHECK_LAUNCH_MODE.format(launch_modes),
        passed=len(launch_modes)
    )


def test_override_launch_mode_with_mark(
    run_test: RunTest, launch_mode_conf: str | None
) -> None:
    run_test(
        """
import pytest

@pytest.mark.script_launch_mode('inprocess')
def test_inprocess(script_runner):
    assert script_runner.launch_mode == 'inprocess'

@pytest.mark.script_launch_mode('subprocess')
def test_subprocess(script_runner):
    assert script_runner.launch_mode == 'subprocess'

@pytest.mark.script_launch_mode('both')
def test_both(script_runner, accumulator=set()):
    assert script_runner.launch_mode not in accumulator
    accumulator.add(script_runner.launch_mode)
        """,
        passed=4,
        launch_mode_conf=launch_mode_conf
    )


def test_help_message(testdir: pytest.Testdir) -> None:
    result = testdir.runpytest(
        '--help',
    )
    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines([
        'console-scripts:',
        '*--script-launch-mode=*',
        '*--hide-run-results*',
    ])
